/***************************************************************************
 *
 * SinglyLinkedList.java
 *
 * An implementation of the List interface using a singly linked list.
 *
 *****************************************************************************/

import java.util.*;

public class SinglyLinkedList<E> implements List<E> {
   	// Remark: All Implemented Interfaces:
    	//java.lang.Iterable<E>, java.util.Collection<E>, java.util.List<E> 


 /*******************************************************
 *
 *  The Node class
 *
 ********************************************************/
   protected static class Node<E> {
      private E data;
      private Node<E> next;

      public Node(E data, Node<E> next) {
         this.data = data;
         this.next = next;
      }
   }  //end-class-Node


   private Node<E> head;
   private int size = 0;


 /**
   *  Constructs an empty list
   */
   public SinglyLinkedList()
   {
      head = null;
   }

 /**
   *  Returns true if the list is empty
   *
   */
   public boolean isEmpty()
   {
      return head == null;
   }

 /**
   *  Returns the length of the list
   *
   */
   public int size()  
   {
      return size;
   }

 /**
   *  Inserts a new node at the beginning of this list.
   *
   */
   public void addFirst(E item)
   {
      head = new Node<E>(item, head);
      size++;	
   }

 /**
   *  Returns the first element in the list.
   *
   */
   public E getFirst()
   {
      if(head == null) throw new NoSuchElementException();

      return head.data;
   }

 /**
   *  Removes the first element in the list.
   *
   */
   public E removeFirst()
   {
      E tmp = getFirst();
      head = head.next;
      size--;
      return tmp;
   }

 /**
   *  Inserts a new node to the end of this list.
   *
   */
   public void addLast(E item)
   {
      if( head == null)
         addFirst(item);
      else
      {
         Node<E> tmp = head;
         while(tmp.next != null) tmp = tmp.next;
         tmp.next = new Node<E>(item, null);
	 size++;
      }
   }


 /**
   *  Returns the last element in the list.
   *
   */
   public E getLast()
   {
      if(head == null) throw new NoSuchElementException();

      Node<E> tmp = head;
      while(tmp.next != null) tmp = tmp.next;
      return tmp.data;
   }


 /**
   *  Removes the last element in the list.
   *
   */
   public E removeLast()
   {  

    if(head == null) throw new NoSuchElementException();
      

      Node<E> prev, cur;
      E elem; 

      if (head.next == null) {
	elem = head.data;
	head = null;
	size--;
	return elem;
      }	
	
      prev = head;
      cur = head.next;	
      while(cur.next != null) { prev = cur; cur = cur.next; }
      elem = cur.data;	
      prev.next = null;
      size--;
      return elem;
   }


 /**
   *  Removes all nodes from the list.
   *
   */
   public void clear()
   {
      head = null; size=0;
   }

/**
   *  Inserts a new node to the end of this list.
   *
   */
   public boolean add(E item)
   {
      this.addLast(item);
      return true;	
   }

/**
   *  Inserts the specified item at the specified position in this list.
   *
   */
   public void add(int index, E item)
   {
      if(index == 0) 
      {
         addFirst(item); 
         return;
      }

      Node<E> cur = head;
      int count = 0;	

      while(cur != null && count < index-1) {
	cur = cur.next;
	count++;
      }

      if(cur != null) {
         cur.next = new Node<E>(item, cur.next);
	 size++;
      } else throw new IndexOutOfBoundsException();
   }




 /**
   *  Returns true if this list contains the specified element.
   *
   */
   public boolean contains(Object x)
   {
      for(E tmp : this)
         if(tmp.equals(x)) return true;

      return false;
   }


 /**
   *  Returns the data at the specified position in the list.
   *
   */
   public E get(int pos)
   {
      if (head == null) throw new IndexOutOfBoundsException();

      Node<E> tmp = head;
      for (int k = 0; k < pos; k++) {
	tmp = tmp.next;
     	if( tmp == null) throw new IndexOutOfBoundsException();
      }
      return tmp.data;
   }


 /**
   *  Returns a string representation
   *
   */
   public String toString()
   {
      StringBuffer result = new StringBuffer();  
      for(Object x : this)
      	result.append(x + " ");

      return result.toString();
   }

 /**
   *  Inserts a new node after a node containing the key.
   *
   */
   public void insertAfter(E key, E toInsert)
   {
      Node<E> tmp = head;

      while(tmp != null && !tmp.data.equals(key)) tmp = tmp.next;

      if(tmp != null) {
         tmp.next = new Node<E>(toInsert, tmp.next);
	 size++;
      }
   }

 /**
   *  Inserts a new node before a node containing the key.
   *
   */
   public void insertBefore(E key, E toInsert)
   {
      if(head == null) return;

      if(head.data.equals(key))
      {
         addFirst(toInsert); size++;
         return;
      }

      Node<E> prev = null;
      Node<E> cur = head;

      while(cur != null && !cur.data.equals(key))
      {
         prev = cur;
         cur = cur.next;
      }
      //insert between cur and prev
      if(cur != null) {
         prev.next = new Node<E>(toInsert, cur);
	 size++;
      }
   }

 /**
   *  Removes the first occurrence of the specified element in this list.
   *
   */
   public boolean remove(Object key)
   {
      if(head == null)
         // throw new RuntimeException("cannot delete");
	return false;

      if( head.data.equals(key) )
      {
         head = head.next;
	 size--;
         return true;
      }

      Node<E> prev = null;
      Node<E> cur  = head;


      while(cur != null && !cur.data.equals(key) )
      {
         prev = cur;
         cur = cur.next;
      }

      if(cur == null)
         // throw new RuntimeException("cannot delete");
	return false;

      //delete cur node
      prev.next = cur.next;
      size--;
      return true;	
   }

   
   

   /*******************************************************
   *
   *  The Iterator class
   *
   ********************************************************/


     public Iterator<E> iterator()
     {
        return new LinkedListIterator();
     }


     private class LinkedListIterator  implements Iterator<E>
     {
        private Node<E> nextNode;

        public LinkedListIterator()
        {
           nextNode = head;
        }

        public boolean hasNext()
        {
           return nextNode != null;
        }

        public E next()
        {
           if (!hasNext()) throw new NoSuchElementException();
           E res = nextNode.data;
           nextNode = nextNode.next;
           return res;
        }

        public void remove() { throw new UnsupportedOperationException(); }

     }  //end class LinkedListIterator





   public E remove(int index)
   //Removes the element at the specified position in this list
   { throw new UnsupportedOperationException(); }

   public E set(int index, E element)
   // Replaces the element at the specified position in this list with the specified element.
   { throw new UnsupportedOperationException(); }

   public List<E> subList(int fromIndex, int toIndex)
   // Returns a view of the portion of this list between the specified fromIndex, inclusive, and toIndex, exclusive. 
   { throw new UnsupportedOperationException(); }

   public int indexOf(Object o)
   // Returns the index of the first occurrence of the specified element in this list,
   // or -1 if this list does not contain the element.
   { throw new UnsupportedOperationException(); }

   public int lastIndexOf(Object o)
   // Returns the index of the last occurrence of the specified element in this list,
   // or -1 if this list does not contain the element.
   { throw new UnsupportedOperationException(); }

   public ListIterator<E> listIterator()
   // Returns a list iterator over the elements in this list (in proper sequence)
   { throw new UnsupportedOperationException(); }

   public ListIterator<E> listIterator(int index)
     	// Returns a list iterator over the elements in this list (in proper sequence),
	// starting at the specified position in the list.
   { throw new UnsupportedOperationException(); }

   public boolean retainAll(Collection<?> c)
   { throw new UnsupportedOperationException(); }

   public boolean removeAll(Collection<?> c)
   { throw new UnsupportedOperationException(); }

   public boolean addAll(Collection<? extends E> c)
   { throw new UnsupportedOperationException(); }

   public boolean addAll(int index, Collection<? extends E> c)
   { throw new UnsupportedOperationException(); }

   public boolean containsAll(Collection<?> c)
   { throw new UnsupportedOperationException(); }

   public <T> T[] toArray(T[] a)
   { throw new UnsupportedOperationException(); }

   public Object[] toArray()
   { throw new UnsupportedOperationException(); }


	/****** ESERCITAZIONE ************************************/
   
   
	 /**
	   *  Reverses the list
	   *  Complexity: O(n)
	   */
	   public SinglyLinkedList<E> reverse()
	   {
	      SinglyLinkedList<E> list = new SinglyLinkedList<E>();
	      Node<E> tmp = head;
	      while(tmp != null)
	      {
	         list.addFirst( tmp.data );
	         tmp = tmp.next;
	      }
	      return list;
	   }


	 /**
	   *  Returns a copy of the immutable list
	   *  Complexity: O(n^2)
	   */
	   public  SinglyLinkedList<E> copy1()
	   {
	      SinglyLinkedList<E> twin = new SinglyLinkedList<E>();
	      Node<E> tmp = head;
	      while(tmp != null)
	      {
	         twin.addLast( tmp.data ); // O(n) in questa implementazione
	         tmp = tmp.next;
	      }

	      return twin;
	   }

	 /**
	   *  Returns a copy of the immutable list
	   *  Complexity: O(n)
	   */
	   public SinglyLinkedList<E> copy2()
	   {
	      SinglyLinkedList<E> twin = new SinglyLinkedList<E>();
	      Node<E> tmp = head;
	      while(tmp != null)
	      {
	         twin.addFirst( tmp.data );
	         tmp = tmp.next;
	      }

	      return twin.reverse();
	   }



	 /**
	   *  Returns a copy of the immutable list
	   *  It uses a tail reference.
	   *  Complexity: O(n)
	   */
	   public SinglyLinkedList<E> copy3()
	   {
	      SinglyLinkedList<E> twin = new SinglyLinkedList<E>();
	      Node<E> tmp = head;
	      if(head==null) return null;
	      twin.head = new Node<E>(head.data, null);  //first node of twin
	      Node<E> tailOfTwin = twin.head;  //tail
	      while(tmp.next != null)
	      {
	         tmp = tmp.next;
	         tailOfTwin.next = new Node<E>(tmp.data, null);
	         tailOfTwin = tailOfTwin.next;
	      }

	      return twin;
	   }



/**
* EX: Completare la classe SinglyLinkedList con un metodo 
* che verifica se la lista corrente  uguale alla lista data
* ovvero se hanno lo stesso contenuto (nello stesso ordine)
*/



 public boolean equals(Object obj) {

 if (this == obj) return true;
 if (obj == null) return false;
 if (!(obj instanceof SinglyLinkedList<?>) ) return false;
 SinglyLinkedList<?> l2 = (SinglyLinkedList<?>) obj;

 Node<E> tmp1 = this.head;
 Node<?> tmp2 = l2.head;
 while(tmp1!= null && tmp2!= null){
	//Confronto dei corrispondenti elementi	
	if(! tmp1.data.equals(tmp2.data)) return false;
	else {
	      tmp1 = tmp1.next;
            tmp2 = tmp2.next;
     	}
 }
 // verifica delle lunghezze delle liste 
 if (tmp1 == null && tmp2 == null) return true;
 else return false;

}


public boolean equalsRec(Object obj) {
	
 if (this == obj) return true;
 if (obj == null) return false;
 if (!(obj instanceof SinglyLinkedList<?>) ) return false;
 return equalsRec(this.head, ((SinglyLinkedList<?>)obj).head);
}


private boolean equalsRec(Node<E> n1, Node<?> n2) {
 // condizioni di terminazione
 if (n1 == null && n2 == null) return true; 
 else if (n1 == null || n2 == null) return false;
 // passo ricorsivo
 if ( ! ((n1.data).equals(n2.data)) )
	return false;
 else return equalsRec(n1.next, n2.next);

}


} // end-class SinglyLInkedList


